/*
     []-------------------------------------------------------------[]
     |		                 BASICO.C                            |
     |						 	             |
     |    Funcoes matematicas (trigonometricas, logaritmica, ...)    |
     |       operacoes e funcoes para calculo vetorial e outras      |
     |						                     |
     |            por Lenimar N. Andrade, maio/97-maio/98            |
     []-------------------------------------------------------------[]
*/

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#if (defined(__MSDOS__) || defined(__TURBOC__))
#include <dos.h>
#include <conio.h>
#undef UNIX
#else
#define UNIX
/* #define MATEMAT */ /* Funcoes matemat. def. em BASICO.C e nao em math.h */
#endif

#ifndef PI
#define PI		3.14159265358979323846
#endif

#define PI_SOBRE_2	1.57079632679489661923
#define PI_SOBRE_4      0.78539816339744830961
#define DOIS_PI         6.28318530717958647692

#define INF     3.4e38
#define IINF      5000

/* Quadrado de x */
#define SQR(x)  ((x)*(x))

/* Distancia entre os pontos (x1, y1, z1) e (x2, y2, z2) */
#define DISTANCIA(x1,y1,z1,x2,y2,z2) sqrt(SQR((x1)-(x2))+SQR((y1)-(y2))+SQR((z1)-(z2)))

/* Determinante da matriz {{a, b}, {c, d}} */
#define Det2x2(a, b, c, d)  ((a)*(d)-(b)*(c))

/* Determinante da matriz {{a, b, c}, {d, e, f}, {g, h, i}} */
#define Det3x3(a,b,c,d,e,f,g,h,i)  ((a)*(e)*(i)+(c)*(d)*(h)+(b)*(f)*(g)-(c)*(e)*(g)-(a)*(f)*(h)-(b)*(d)*(i))

/* Numero maximo de iteracoes permitido */
#define MAX_ITERACOES 100

/* f'(a) = derivada de f no ponto a */
#define DER(f,a,h)  ((f(a+h)-f(a-h))/(2.0*h))

/* f"(a) = derivada segunda de f no ponto a */
#define DER2(f,a,h) ((-2.0*f(a)+f(a-2.0*h)+f(a+2.0*h))/(4.0*h*h))

/* f'''(a) = derivada terceira de f no ponto a */
#define DER3(f,a,h) ((-f(a-3*h)+3*f(a-h)-3*f(a+h)+f(a+3*h))/(8.0*h*h*h))

/* Derivada parcial de f com relacao `a primeira variavel */
#define DIF1(f,a,b,h)  ((f(a+h,b)-f(a-h,b))/(2.0*h))

/* Derivada parcial de f com relacao `a segunda variavel */
#define DIF2(f,a,b,h)  ((f(a,b+h)-f(a,b-h))/(2.0*h))

#define MOSTRA_MENS_ERRO if (houve_erro) printf(mens_erro)

/* Variaveis globais */
double h_erro = 1e-6,      /* Usado no calculo aproximado das derivadas    */
       z_erro = 1e-6,      /* Se |x| < z_erro, entao x sera' considerado 0 */
       r_erro = 0.01;      /* Usado para verificar retorno ponto inicial   */

char mens_erro[100];       /* Mensagem de erro                             */
int houve_erro = 0;        /* Determina se houve erro na execucao          */

char NOME_ARQ_ERROS[30] = "erros.msg";

typedef struct { double x, y, z, X, Y, Z; } caixa;
typedef struct { double xmin, xmax, ymin, ymax; } retangulo;
#define Dominio retangulo

/* Produto vetorial (v1, v2, v3) = (a1, a2, a3) x (b1, b2, b3) */
#define PROD_VET(v1,v2,v3,a1,a2,a3,b1,b2,b3)  v1=(a2)*(b3)-(a3)*(b2);v2=(a3)*(b1)-(a1)*(b3);v3=(a1)*(b2)-(a2)*(b1)

/* Norma do vetor (v1, v2, v3) */
#define NORMA(v1, v2, v3)       (sqrt((v1)*(v1)+(v2)*(v2)+(v3)*(v3)))

typedef struct { double x, y, z; } vetor;       /* Vetor (x, y, z) no R3   */
typedef struct { double x, y, z; } ponto;       /* Ponto (x, y, z) no R3   */
typedef struct { double u, v; } ponto_dominio;  /* Ponto (u, v) no dominio */

/* ----------------------------------------------------------------------- */

#ifdef MATEMAT

#define  cos(x)   Cos(x)
#define  sin(x)   Sen(x)
#define  cosh(x)  Cosh(x)
#define  sinh(x)  Senh(x)
#define  fabs(x)  Modulo(x)
#define  atan(x)  ArcTg(x)
#define  asin(x)  ArcSen(x)
#define  acos(x)  ArcCos(x)
#define  log(x)   Ln(x)
#define  ln(x)    Ln(x)
#define  exp(x)   Exp(x)
#define  sqrt(x)  Sqrt(x)
#define  pow(x,y) Pot(x,y)
#define  tan(x)   (Sen(x)/Cos(x))
#define  ceil(x)  Ceil(x)
#define  floor(x) Floor(x)
#define  Tg(x)    (Sen(x)/Cos(x))

/* ----------------------------------------------------------------------- */

double Cos(double x);

double Sen(double x) {

/* Calcula o seno de x. E' usada a serie de Taylor

	     Sen(x) = x - x^3/3! + x^5/5! - x^7/7! + ...

   e diversas formulas de reducao ao primeiro quadrante */

  double termo = 1, soma = 0, sinal = 1.0, erro = 1e-16;
  int i = 1;

  if (x < 0) { x = -x; sinal = -1.0; }
  while (x >= DOIS_PI) x -= DOIS_PI;

  if (x > PI && x < DOIS_PI) { sinal = -sinal; x = x - PI; }

  if (PI_SOBRE_2 < x && x <= PI) x = PI - x;

  if (x == 0) return 0.0;
  else if (x == PI_SOBRE_2) return sinal;

  if (PI_SOBRE_4 < x && x < PI_SOBRE_2) return sinal*Cos(PI_SOBRE_2 - x);

  do {
    termo *= x/i;
    if (i % 2) { if (i % 4 == 1) soma += termo; else soma -= termo; }
    i++;
  } while (termo >= erro);

  return sinal*soma;
}

/* ----------------------------------------------------------------------- */

double Cos(double x) {

/* Calcula o cosseno de x. E' usada a serie de Taylor

	    Cos(x) = 1 - x^2/2! + x^4/4! - x^6/6! + ...

   e diversas formulas de reducao ao primeiro quadrante */

  double termo = 1, soma = 1, erro = 1e-16, sinal = 1.0;
  int i = 1;

  if (x < 0) x = -x;
  while (x >= DOIS_PI) x -= DOIS_PI;

  if (PI < x && x < DOIS_PI) x = DOIS_PI - x;

  if (PI_SOBRE_2 < x && x < PI) { x = PI - x; sinal = -1.0; }

  if (PI_SOBRE_4 < x && x < PI_SOBRE_2) return sinal*Sen(PI_SOBRE_2 - x);

  if (x == 0) return sinal;
  else if (x == PI_SOBRE_2) return 0;

  do {
    termo *= x/i;
    if (i % 2 == 0) { if (i % 4 == 2) soma -= termo; else soma += termo; }
    i++;
  } while (termo >= erro);

  return sinal*soma;
}

/* ----------------------------------------------------------------------- */

double Modulo(double x) {      /* Calcula |x| */

  if (x >= 0) return x; else return -x;

}

/* ----------------------------------------------------------------------- */

double Exp(double x) {

/* Calcula exponencial de base e = 2.718... de x. E' usada a serie
   de Taylor

	   Exp(x) = 1 + x + x^2/2! + x^3/3! + x^4/4! + ...

   E' usado tambem que  Exp(x) = Exp(x/p) * ... * Exp(x/p)  (p fatores) */

  unsigned int i = 1;
  unsigned long p, k;
  double soma = 1, termo = 1, erro = 1e-16;
  double z, prod = 1;

  if (Modulo(x) > 1) {
    p = 1 + (unsigned long) Modulo(x);
    z = Exp(x/p);
    for (k = 1; k <= p; k++) prod *= z;
    return prod;
  }

  do {
    termo *= x/i;
    i++;
    soma += termo;
  } while (Modulo(termo) >= erro);

  return soma;
}


/* ----------------------------------------------------------------------- */

double Cosh(double x) {      /* Cosseno hiperbolico de x */

  return 0.5*(Exp(x)+Exp(-x));

}

/* ----------------------------------------------------------------------- */

double Sinh(double x) {      /* Seno hiperbolico de x */

  return 0.5*(Exp(x)-Exp(-x));

}

/* ----------------------------------------------------------------------- */

double Modulo(double x) {      /* Calcula |x| */

  if (x >= 0) return x; else return -x;

}

/* ----------------------------------------------------------------------- */

#define  EXP1       2.71828182845904523536
#define  EXP2       7.38905609893065022723
#define  EXP3      20.08553692318766774092
#define  EXP4      54.59815003314423907811
#define  EXP5     148.41315910257660342111
#define  EXP6     403.42879349273512260838
#define  EXP7    1096.63315842845859926372
#define  EXP8    2980.95798704172827474359
#define  EXP9    8103.08392757538400770999
#define  EXP10  22026.46579480671651695790

double Ln(double x) {

/* Calcula o logaritmo natural de x, usando a serie de Taylor

   1/2 Ln((1 + x)/(1 - x)) = x + x^3/3 + x^5/5 + x^7/7 + ...

   E' usado tambem  que se x > e^n entao Ln(x) = n + Ln(x/e^n) , e=2.718... */

  int i = 1;
  double y, soma, pot_y, termo, erro = 1e-16;

  if (x > EXP10) return (10.0 + Ln(x/EXP10));
  else if (x > EXP9) return (9.0 + Ln(x/EXP9));
  else if (x > EXP8) return (8.0 + Ln(x/EXP8));
  else if (x > EXP7) return (7.0 + Ln(x/EXP7));
  else if (x > EXP6) return (6.0 + Ln(x/EXP6));
  else if (x > EXP5) return (5.0 + Ln(x/EXP5));
  else if (x > EXP4) return (4.0 + Ln(x/EXP4));
  else if (x > EXP3) return (3.0 + Ln(x/EXP3));
  else if (x > EXP2) return (2.0 + Ln(x/EXP2));
  else if (x > EXP1) return (1.0 + Ln(x/EXP1));

  y = (x - 1.0)/(x + 1.0);
  soma = pot_y = y;
  do {
    pot_y *= y*y;
    termo = pot_y/(2.0*i + 1.0);
    soma += termo;
    i++;
  } while (Modulo(termo) >= erro);

  return (2.0*soma);
}

/* ----------------------------------------------------------------------- */

double Pot(double base, double expoente) {

/* Calcula a potencia base^expoente */

  double z;

  z = Ln(base);
  return Exp(expoente*z);
}

/* ----------------------------------------------------------------------- */

double Sqrt(double x) {

/* Calcula a raiz quadrada de x, calculando o limite da sequencia

		    a     =  0.5 ( a  + x/a )
		     n+1            n      n                               */

  double a_n = 1.0, a_n_mais_1 = 1.0, erro = 1e-16;

  if (x <= 0) return 0.0;

  do {
    a_n = a_n_mais_1;
    a_n_mais_1 = 0.5*(a_n + x/a_n);
  } while (Modulo(a_n_mais_1 - a_n) >= erro);

  return a_n_mais_1;
}

/* ----------------------------------------------------------------------- */

double ArcSen(double x) {

  double termo, soma = 0.0, erro = 1e-16;
  long n = 1;


  if (Modulo(x) > 1) return -2;
  if (x == 1) return PI_SOBRE_2;
  else if (x == -1) return -PI_SOBRE_2;

  soma = termo = x, n = 1;
  do {
    termo *= ((2.0*n - 1.0)/(2.0*n + 1.0))*((2.0*n - 1.0)/(2.0*n))*x*x;
    soma += termo;
    n++;
  } while (Modulo(termo) >= erro);

  return soma;

}

/* ----------------------------------------------------------------------- */

double ArcCos(double x) {

  if (Modulo(x) > 1) return -1;
  return PI_SOBRE_2 - ArcSen(x);

}

/* ----------------------------------------------------------------------- */

double ArcTg(double x) {

  double acrescimo, soma, termo, erro = 1e-16;
  int n, sinal;

  if (x == 0) return 0;
  else if (x == 1) return PI_SOBRE_4;
  else if (x == -1) return -PI_SOBRE_4;
  else if (x > 1) return PI_SOBRE_2 - ArcTg(1/x);
  else if (x < -1) return -PI_SOBRE_2 - ArcTg(1/x);

  soma = termo = x, n = 1;
  do {
    if (n % 2 == 0) sinal = 1; else sinal = -1;
    termo *= x*x;
    acrescimo = sinal*termo/(2*n + 1);
    soma += acrescimo;
    n++;
  } while (Modulo(acrescimo) >= erro);

  return soma;

}


/* ----------------------------------------------------------------------- */

double MaiorInteiro(double x) {

 /* Calcula [x] = n, se n e'inteiro tal que n <= x < n + 1 */

  long p;

  p = (long) x;
  if (p <= x) return (double) p; else return (double) (p - 1);
}

/* ----------------------------------------------------------------------- */

double Floor(double x) {

  long p;

  p = (long) x;
  return (double) p;
}

/* ----------------------------------------------------------------------- */

double Ceil(double x) {

  long p;

  p = (long) x;
  return (double) (1.0 + p);
}

#endif

/* ----------------------------------------------------------------------- */

void Erro(char *mens_erro) {

/* Salva mensagem de erro em arquivo */

  FILE *arq_erros;

  arq_erros = fopen(NOME_ARQ_ERROS, "at");
  fprintf(arq_erros, mens_erro);
  houve_erro = 1;
  fclose(arq_erros);
}

/* ----------------------------------------------------------------------- */

int ResolveSistema3x3(double A, double B, double C, double M,
		      double D, double E, double F, double N,
		      double G, double H, double I, double P,
		      double *X, double *Y, double *Z) {

/* Resolucao do sistema linear determinado
			 Ax + By + Cz = M
			 Dx + Ey + Fz = N
			 Gx + Hy + Iz = P
   Devolve 0 somente se a solucao unica nao puder ser encontrada. */

  double det;

  det = Det3x3(A, B, C, D, E, F, G, H, I);
  if (fabs(det) < z_erro) return 0;
  *X = Det3x3(M, B, C, N, E, F, P, H, I)/det;
  *Y = Det3x3(A, M, C, D, N, F, G, P, I)/det;
  *Z = Det3x3(A, B, M, D, E, N, G, H, P)/det;
  return 1;
}

/* ------------------------------------------------------------------------ */

int VetorNormal(vetor *N, double (*f1)(double,double),
		double (*f2)(double, double), double (*f3)(double, double),
		double u0, double v0) {

/* Calcula N, o vetor normal no  ponto  (u0, v0)  `a
   superficie f(u, v) = (f1(u, v), f2(u, v), f3(u, v)). Retorna 0
   se o vetor normal encontrado for nulo ou 1 em  caso contrario. */

  double a1, a2, a3, b1, b2, b3, norma_N;

  /* Calculo de 2 vetores tangentes `a superficie f */

  a1 = DIF1(f1, u0, v0, h_erro);
  a2 = DIF1(f2, u0, v0, h_erro);
  a3 = DIF1(f3, u0, v0, h_erro);

  b1 = DIF2(f1, u0, v0, h_erro);
  b2 = DIF2(f2, u0, v0, h_erro);
  b3 = DIF2(f3, u0, v0, h_erro);

  /* Vetor normal a f */
  PROD_VET(N->x, N->y, N->z, a1, a2, a3, b1, b2, b3);

  norma_N = NORMA(N->x, N->y, N->z);
  if (norma_N < z_erro) return 0;
  return 1;
}

/* ----------------------------------------------------------------------- */

int VetorTangente(vetor *T, double (*f1)(double, double),
		   double (*f2)(double, double), double (*f3)(double, double),
		   ponto_dominio pF,
		   double (*g1)(double, double), double (*g2)(double, double),
		   double (*g3)(double, double), ponto_dominio pG) {

/* Calcula T, o vetor unitario tangente `a curva intersecao das superficies
	       f(u, v) = (f1(u, v), f2(u, v), f3(u, v))  e
	       g(u, v) = (g1(u, v), g2(u, v), g3(u, v))
   Retorna 0 somente se houver erro no calculo do vetor tangente. */

  double a1, a2, a3, b1, b2, b3, c1, c2, c3, d1, d2, d3,
	F_n1, F_n2, F_n3, G_n1, G_n2, G_n3, norma_T;

  /* Verifica se o ponto pertence mesmo `a intersecao */
  if (DISTANCIA(f1(pF.u, pF.v), f2(pF.u, pF.v), f3(pF.u, pF.v),
	g1(pG.u, pG.v), g2(pG.u, pG.v), g3(pG.u, pG.v)) > z_erro) return 0;

  /* Calculo de 2 vetores tangentes `a superficie f */

  a1 = DIF1(f1, pF.u, pF.v, h_erro);
  a2 = DIF1(f2, pF.u, pF.v, h_erro);
  a3 = DIF1(f3, pF.u, pF.v, h_erro);

  b1 = DIF2(f1, pF.u, pF.v, h_erro);
  b2 = DIF2(f2, pF.u, pF.v, h_erro);
  b3 = DIF2(f3, pF.u, pF.v, h_erro);

  /* Calculo de 2 vetores tangentes `a superficie g */

  c1 = DIF1(g1, pG.u, pG.v, h_erro);
  c2 = DIF1(g2, pG.u, pG.v, h_erro);
  c3 = DIF1(g3, pG.u, pG.v, h_erro);

  d1 = DIF2(g1, pG.u, pG.v, h_erro);
  d2 = DIF2(g2, pG.u, pG.v, h_erro);
  d3 = DIF2(g3, pG.u, pG.v, h_erro);

  /* Vetor normal a f */
  PROD_VET(F_n1, F_n2, F_n3, a1, a2, a3, b1, b2, b3);

  /* Vetor normal a g */
  PROD_VET(G_n1, G_n2, G_n3, c1, c2, c3, d1, d2, d3);

  /* Vetor tangente `a intersecao */
  PROD_VET(T->x, T->y, T->z, F_n1, F_n2, F_n3, G_n1, G_n2, G_n3);

  norma_T = NORMA(T->x, T->y, T->z);
  if (norma_T < z_erro) return 0;

  T->x /= norma_T;
  T->y /= norma_T;
  T->z /= norma_T;

  return 1;
}

/* ----------------------------------------------------------------------- */

void ProximoPontoVetorTangente(ponto *R, vetor T, double L) {

/* Determina o ponto distante L unidades de R na direcao do vetor
   tangente unitario T */

  double norma;

  norma = NORMA(T.x, T.y, T.z);
  L = L/norma;
  R->x += L*T.x;
  R->y += L*T.y;
  R->z += L*T.z;
}

/* ----------------------------------------------------------------------- */

#ifndef UNIX

void HoraAtual(struct time *hora) {

  gettime(hora);
}

/* ----------------------------------------------------------------------- */

double DifTempo(struct time inicio, struct time fim) {

  double tempo_inicio, tempo_fim;

  tempo_inicio = inicio.ti_hour*3600.0 + inicio.ti_min*60.0 +
		 inicio.ti_sec*1.0 + inicio.ti_hund/100.0;
  tempo_fim = fim.ti_hour*3600.0 + fim.ti_min*60.0 +
	      fim.ti_sec*1.0 + fim.ti_hund/100.0;

  return (tempo_fim - tempo_inicio);
}

/* ----------------------------------------------------------------------- */

void CriaArquivoTempoGasto(struct time inicio, struct time fim, char *mens,
                                                             char *NomeArq) {

  double tempo_gasto;
  FILE *arq;
  static int vez = 1;

  if (vez == 1) {
    if ((arq = fopen(NomeArq, "wt")) == NULL) {
      printf("\n\nERRO na criacao do arquivo %s\n", NomeArq);
      return;
    }
    vez++;
  } else {
    if ((arq = fopen(NomeArq, "at")) == NULL) {
      printf("\n\nERRO na abertura do arquivo %s\n", NomeArq);
      return;
    }
  }

  fprintf(arq, "\n%s\n", mens);
  tempo_gasto = DifTempo(inicio, fim);
  fprintf(arq, "\nInicio `as  %d h %d min %d s %d\n", inicio.ti_hour,
			 inicio.ti_min, inicio.ti_sec, inicio.ti_hund);
  fprintf(arq, "\nFim    `as  %d h %d min %d s %d\n", fim.ti_hour,
				 fim.ti_min, fim.ti_sec, fim.ti_hund);
  fprintf(arq, "\nTempo total gasto = %11.5f segundos\n", tempo_gasto);
  fclose(arq);
}

/* ----------------------------------------------------------------------- */

#endif

/*** FIM DE BASICO.C ***/
