/*
     []-------------------------------------------------------------[]
     |		              CONVERG.C                              |
     |						 	             |
     |     Determinacao de  sequencias que convergem para pontos     |
     |     sobre uma intersecao de duas superficies parametricas     |
     |						                     |
     |           por Lenimar N. Andrade, maio/96-abril/97            |
     []-------------------------------------------------------------[]
*/

int DOMINIO_LIMITADO = 1, /* Define se os dominios sao limitados         */
    ARQ_VETORES = 1,      /* Define se salva vetores no arquivo          */
    ARQ_DOMINIO = 1,      /* Define se salva pontos do dominio           */
    ARQ_CIRC  = 0,        /* Def. se mostra raio e centro das circunfer. */
    ITERACOES = 1,        /* Define se mostra numero de iteracoes        */
    ARQ_DEC = 6,          /* Numero de casas decimais usadas no arquivo  */
    METRICA = 1,          /* Metrica usada na identificacao de pontos    */
    SOM = 0,              /* Define se emite som no final da execucao    */
    CIRCULAR = 1,         /* Define se usa o metodo do passo circular    */
    DOMINIO2 = 0,         /* Distribui pontos tambem no segundo dominio  */
    NIVEL_SUBDIV = 3,     /* Nivel maximo de subdivisao dos dominios     */
    NIVEL_CLASS = 2,      /* Nivel de classificacao de pts. encontrados  */
    n_PONTOS = IINF,      /* Maximo de pontos desenhados em um trecho    */
    ISOLADO = 0,          /* Verifica a existencia de pontos isolados    */
    SALVA_HIST = 1,       /* Registra etapas da execucao em arquivo      */
    MOSTRA_HIST =1,       /* Registra etapas da execucao na tela         */
    INVERSO = 1,          /* Percorre trecho no sentido contrario        */
    INICIAL = 1,          /* Determina se marca ponto inicial/final      */
    VERIFICA_RETORNO = 1, /* Verifica se houve retorno ao ponto inicial  */
    GRADE = 1,            /* Se for > 1, desenha grade nos dominios      */
    DESENHAR_PONTOS_DOM=1;/* Determina se desenha pontos nos dominios    */

FILE *arq_padrao,  /* Arquivos dos pontos e vetores tangentes `a inters. */
     *arq_execucao;  /* Arquivo que registra as etapas da execucao       */

char ARQ_DADOS_PADRAO[30] = "dados.txt", /* Nome padrao do arq. de dados */
     ARQ_DADOS_DISTINTOS[30] = "dados2.txt",    /* Arq. pontos distintos */
     ARQ_TEMPO[30] = "tempo.txt", /* Arquivo com tempo gasto na execucao */
     ARQ_EXECUCAO[30] = "execucao.txt"; /* Historia da execucao          */

/* Variaveis globais usadas para desenhar pontos nos dominios */
int a_domF = 10, b_domF = 10, c_domF = 160, d_domF = 160,
    cor_pontos_domF = 3, cor_molduraF = 8,
    a_domG = 10, b_domG = 250, c_domG = 160, d_domG = 400,
    cor_pontos_domG = 4, cor_molduraG = 8;

double PASSO_L = 0.05,  /* Comprimento do passo da caminhada            */
      DISTANCIA = 0.1;  /* Distancia minima entre os pontos encontrados */

char ARQ_CONFIGURACAO[100], CONFIG_TXT[30] = "config.txt";

typedef struct {float x, y, z, t; } ponto_R4;

#define MAX_DISTINTOS 3000
#define DIST1(x1,x2,x3,x4,y1,y2,y3,y4)  fabs((x1)-(y1))+fabs((x2)-(y2))+fabs((x3)-(y3))+fabs((x4)-(y4))
#define DIST2(x1,x2,x3,x4,y1,y2,y3,y4)  sqrt(SQR((x1)-(y1))+SQR((x2)-(y2))+SQR((x3)-(y3))+SQR((x4)-(y4)))
#define DIST3(x1,x2,y1,y2)  sqrt(SQR((x1)-(y1))+SQR((x2)-(y2)))

/* Constantes usadas somente na construcao de graficos na tela */
int X_CENTRO = 319, Y_CENTRO = 239,  /* Coordenadas da origem na tela     */
    TAM = 50,                        /* Constante de ampliacao do grafico */
    LINHA_PONTILHADA = 1,            /* Define se usa linha pontilhada    */
    corX = 14,                 /* Cor de ponto associado no outro dominio */
    corX3D = 15,               /* Cor de ponto associado no espaco R3     */
    X_PONTOS_SUBDIVISAO = 1,   /* Marca com "x" pts encontrados subdiv.   */
    X_OUTROS = 0,
    DES_EIXOS = 0, TAM_EIXOS = 4, COR_EIXOS = 8, COR_GRAF1 = 3,
    COR_GRAF2 = 4, COR_FUNDO = 0, X_COR_DOMINIO = 8,
    nuF = 20, nuG = 20,       /* Subdivisoes usadas para construir o      */
    nvF = 20, nvG = 20,       /* graficos das superficies F e G.          */
    USAR_MOUSE = 1,        /* Define se usa o mouse nos pontos do dominio */
    COR_INTER = 14,        /* Cor da curva-intersecao                     */
    DEMORA = 0,            /* Milisegundos entre dois pontos mostrados    */
    GRAF_X = 1;            /* Define se pontos desenhados em forma de "x" */

double FI = -PI/4, TETA = -PI/4;     /* Angulos de rotacao do grafico 3D  */

char mens_Dom_F[30] = "Dominio_de_F", mens_Dom_G[30] = "Dominio_de_G";

/* ----------------------------------------------------------------------- */

double Angulo(double x, double y) {

/* Retorna o angulo no intervalo [0, 2] que o vetor (x, y) forma
   com (1, 0) */

  double r;

  r = sqrt(x*x + y*y);
  if (r <= z_erro) return 0;
  else if (y >= 0) return acos(x/r);
  else return acos(-x/r) + PI;
}

/* ----------------------------------------------------------------------- */

int PontoNaCircunferencia(ponto *R, ponto P, vetor v1, ponto Q, vetor v2,
						     ponto C, double teta) {

/* Determina o ponto  R na circunferencia no espaco R3 de centro C e que
   passa pelo ponto Q e contida no plano paralelo aos vetores v1 e v2. R
   e' definido como sendo o ponto desta circunferencia tal que o arco QR
   mede  teta radianos (a partir de Q, no sentido de P  para Q). Retorna
   0 somente se houver erro. */

  double r, rot[3][3], beta, a, b, c, p, q, t, lambda, V, lambdaV;
  ponto P2, Q2, R2;

  PROD_VET(a, b, c, v1.x, v1.y, v1.z, v2.x, v2.y, v2.z);

  q = a*v1.x + b*v1.y + c*v1.z;
  if (fabs(q) > z_erro) {
    p = a*(Q.x - P.x) + b*(Q.y - P.y) + c*(Q.z - P.z);
    t = p/q;
    P.x += t*v1.x, P.y += t*v1.y, P.z += t*v1.z;
  }

  r = DISTANCIA(C.x, C.y, C.z, Q.x, Q.y, Q.z);  /* raio da circunferencia */

  /* Calculo da matriz de rotacao rot[][] */
  V = NORMA(a, b, c);
  if (V < z_erro) return 0;
  lambda = sqrt(b*b + c*c);
  lambdaV = lambda*V;
  if (lambda > z_erro) {
    rot[0][0] = lambda/V, rot[0][1] = 0, rot[0][2] = a/V;
    rot[1][0] = -a*b/lambdaV, rot[1][1] = c/lambda, rot[1][2] = b/V;
    rot[2][0] = -a*c/lambdaV, rot[2][1] = -b/lambda, rot[2][2] = c/V;
  } else {
    rot[0][0] = 0, rot[0][1] = 0, rot[0][2] = (a >= 0 ? 1 : -1);
    rot[1][0] = 0, rot[1][1] = 1, rot[1][2] = 0;
    rot[2][0] = (a >= 0 ? -1 : 1), rot[2][1] = 0, rot[2][2] = 0;
  }

  /* Rotacao dos pontos P e Q, segundo a matriz rot[][]. P e Q sao
     levados para P2 e Q2, respectivamente. */
  P2.x = (P.x - C.x)*rot[0][0] + (P.y - C.y)*rot[1][0] + (P.z - C.z)*rot[2][0];
  P2.y = (P.x - C.x)*rot[0][1] + (P.y - C.y)*rot[1][1] + (P.z - C.z)*rot[2][1];
  P2.z = (P.x - C.x)*rot[0][2] + (P.y - C.y)*rot[1][2] + (P.z - C.z)*rot[2][2];

  Q2.x = (Q.x - C.x)*rot[0][0] + (Q.y - C.y)*rot[1][0] + (Q.z - C.z)*rot[2][0];
  Q2.y = (Q.x - C.x)*rot[0][1] + (Q.y - C.y)*rot[1][1] + (Q.z - C.z)*rot[2][1];
  Q2.z = (Q.x - C.x)*rot[0][2] + (Q.y - C.y)*rot[1][2] + (Q.z - C.z)*rot[2][2];

  beta = Angulo(Q2.x, Q2.y);

  /* Calculo do ponto R2 */
  if (P2.x*Q2.y - P2.y*Q2.x > 0) {
    R2.x = r*cos(beta + teta);
    R2.y = r*sin(beta + teta);
    R2.z = 0;
  } else {
    R2.x = r*cos(beta - teta);
    R2.y = r*sin(beta - teta);
    R2.z = 0;
  }

  /* Rotacao inversa do ponto R2 */
  R->x = (R2.x)*rot[0][0] + (R2.y)*rot[0][1] + (R2.z)*rot[0][2] + C.x;
  R->y = (R2.x)*rot[1][0] + (R2.y)*rot[1][1] + (R2.z)*rot[1][2] + C.y;
  R->z = (R2.x)*rot[2][0] + (R2.y)*rot[2][1] + (R2.z)*rot[2][2] + C.z;

  return 1;
}

/* ----------------------------------------------------------------------- */

int CentroDaCircunferencia(ponto *C, double *raio, ponto P, vetor v1,
						   ponto Q, vetor v2) {

  /* Determina  o  centro  da  circunferencia  que  passa  pelos pontos P e
     Q,  cujo centro  situa-se na intersecao dos  planos   que  passam  por
     P  e  Q  e  cujos vetores normais sao v1 e v2. Retorna 0
     somente se houver erro. */

  vetor n;

  PROD_VET(n.x, n.y, n.z, v1.x, v1.y, v1.z, v2.x, v2.y, v2.z);

  if (!ResolveSistema3x3(v1.x, v1.y, v1.z, P.x*v1.x + P.y*v1.y + P.z*v1.z,
			 v2.x, v2.y, v2.z, Q.x*v2.x + Q.y*v2.y + Q.z*v2.z,
			 n.x,  n.y,  n.z,  Q.x*n.x  + Q.y*n.y  + Q.z*n.z,
			 &C->x, &C->y, &C->z)) {
			 *raio = C->x = C->y = C->z = INF; return 0;}

  *raio = DISTANCIA(C->x, C->y, C->z, Q.x, Q.y, Q.z);

  return 1;
}

/* ----------------------------------------------------------------------- */

int ProximoPontoCirc(ponto *R, ponto *C, double *raio,
		     ponto P, vetor v1, ponto Q, vetor v2, double L) {

  if (!CentroDaCircunferencia(C, raio, P, v1, Q, v2)) {
    return 0;
  }

  if (*raio > 1) {
    if (!PontoNaCircunferencia(R, P, v1, Q, v2, *C, L/(*raio))) return 0;
  } else {
    if (!PontoNaCircunferencia(R, P, v1, Q, v2, *C, L)) return 0;
  }

  return 1;
}

/* ----------------------------------------------------------------------- */

int Calcula_P_i_mais_1(ponto P, double (*F)(double, double),
		    double (*G)(double, double), double (*H)(double, double),
		    ponto_dominio P_i, ponto_dominio *P_i_mais_1) {

/* Calcula o proximo elemento da sequencia que converge para um ponto sobre
   a superficie (F(u,v), G(u,v), H(u,v)). Retorna 0 somente se houver erro. */

   double a, b, c, d, e, f, A, B, C, D, detM,
	  a2, b2, c2, d2, e2, f2, A2, B2, C2, D2;

   /*			                     [ a b ]
   J = Matriz 3 x 2 jacobiana de (F, G, H) = [ c d ]
					     [ e f ] */
   a = DIF1(F, P_i.u, P_i.v, h_erro);
   b = DIF2(F, P_i.u, P_i.v, h_erro);
   c = DIF1(G, P_i.u, P_i.v, h_erro);
   d = DIF2(G, P_i.u, P_i.v, h_erro);
   e = DIF1(H, P_i.u, P_i.v, h_erro);
   f = DIF2(H, P_i.u, P_i.v, h_erro);

   /* [ A B ]  =     t
      [ C D ]       J J  */
   A = a*a + c*c + e*e;
   B = C = a*b + c*d + e*f;
   D = b*b + d*d + f*f;
   detM = A*D - B*C;
   if (detM < z_erro) {
     sprintf(mens_erro, "\n\nO vetor (%6.6f, %6.6f, %6.6f) tangente a F"
     " em (%6.6f, %6.6f)", a, c, e, P_i.u, P_i.v);
     Erro(mens_erro);
     sprintf(mens_erro, "\ne o vetor (%6.6f, %6.6f, %6.6f) tangente a F"
     " em (%6.6f, %6.6f)\nsao paralelos", b, d, f, P_i.u, P_i.v);
     Erro(mens_erro);
     Erro("\nERRO: no calculo da funcao Calcula_P_i_mais_1()");
     return 0;
   }

   /*		           -1
	[ A2 B2 ] = [ A B ]    = 1/detM [  D -B ]
	[ C2 D2 ]   [ C D ]             [ -C  A ]  */

   A2 = D/detM;
   B2 = -B/detM;
   C2 = -C/detM;
   D2 = A/detM;

   /*  [ A2 B2 ]   t      [ a2 c2 e2 ]
       [ C2 D2 ]  J    =  [ b2 d2 f2 ]  */

   a2 = a*A2 + b*B2;
   b2 = a*C2 + b*D2;
   c2 = c*A2 + d*B2;
   d2 = c*C2 + d*D2;
   e2 = e*A2 + f*B2;
   f2 = e*C2 + f*D2;

  /*                   [ a2 c2 e2 ] [ P.x - F(P_i) ]
       P_i_mais_1  =   [ b2 d2 f2 ] [ P.y - G(P_i) ] + P_i
		       [          ] [ P.z - H(P_i) ]            */

  P_i_mais_1->u = a2*(P.x - F(P_i.u, P_i.v)) + c2*(P.y - G(P_i.u, P_i.v)) +
				     e2*(P.z - H(P_i.u, P_i.v)) + P_i.u;
  P_i_mais_1->v = b2*(P.x - F(P_i.u, P_i.v)) + d2*(P.y - G(P_i.u, P_i.v)) +
				     f2*(P.z - H(P_i.u, P_i.v)) + P_i.v;

  return 1;
}

/* ----------------------------------------------------------------------- */

ponto_dominio PontoNoDominio(ponto P, double (*F)(double, double),
		    double (*G)(double, double), double (*H)(double, double),
		    ponto_dominio P_i) {

/* Determina o ponto da superficie (F(u,v), G(u,v), H(u,v)) que  esta'
   mais proximo do ponto P. Retorna {INF, INF} somente se houver erro. */

  ponto_dominio Q;
  unsigned cont = 0;

  Q = P_i;
  do {
    P_i = Q;
    if (!Calcula_P_i_mais_1(P, F, G, H, P_i, &Q)) {
      Q.u = INF, Q.v = INF;
      return Q;
    }
    cont++;
  } while ((fabs(Q.u - P_i.u) > z_erro || fabs(Q.v - P_i.v) > z_erro)
						     && cont < MAX_ITERACOES);

  if (cont >= MAX_ITERACOES) { Q.u = INF, Q.v = INF; }

  return Q;
}

/* ----------------------------------------------------------------------- */

int PontoNaIntersecao(double (*f1)(double,double), double (*f2)(double,double),
		   double (*f3)(double,double), ponto_dominio P,
		   double (*g1)(double,double), double (*g2)(double,double),
		   double (*g3)(double,double), ponto_dominio Q) {

  /* Verifica se f1(P.u, P.v) = g1(Q.u, Q.v), f2(P.u, P.v) = g2(Q.u, Q.v)),
     f3(P.u, P.v) = g3(Q.u, Q.v), dentro de uma certa tolerancia. */

  return (fabs(f1(P.u, P.v) - g1(Q.u, Q.v)) < z_erro &&
	  fabs(f2(P.u, P.v) - g2(Q.u, Q.v)) < z_erro &&
	  fabs(f3(P.u, P.v) - g3(Q.u, Q.v)) < z_erro);
}

/* ----------------------------------------------------------------------- */

int DeterminaPontoDaIntersecao(ponto P, ponto_dominio *P0, ponto_dominio *Q0,
		  double (*F1)(double,double), double (*F2)(double,double),
		  double (*F3)(double,double), double (*G1)(double,double),
		  double (*G2)(double,double), double (*G3)(double,double)) {

  /* Determina ponto na intersecao das superficies (F1,F2,F3) e (G1,G2,G3).
     Retorna 0 se houver erro; caso contrario, retorna o numero de iteracoes
     utilizadas. E' usado o ponto P como ponto inicial. */

  ponto P_anterior, P1, P2, n3, P3;
  ponto_dominio p1, q1, p0, q0;
  int num_it = 0;
  double sol1, sol2, sol3;
  vetor n1, n2;

  p0 = *P0, q0 = *Q0;

  do {
    p1 = PontoNoDominio(P, F1, F2, F3, p0);
    if (p1.u == INF || p1.v == INF) {
      sprintf(mens_erro, "\n\nPonto no dominio de F nao encontrado quando "
			     "P = (%6.6f, %6.6f, %6.6f)", P.x, P.y, P.z);
      Erro(mens_erro);
      Erro("\nERRO: no calculo da funcao DeterminaPontoDaIntersecao()");
      return 0;
    }
    q1 = PontoNoDominio(P, G1, G2, G3, q0);
    if (q1.u == INF || q1.v == INF) {
      sprintf(mens_erro, "\n\nPonto no dominio de G nao encontrado quando "
			     "P = (%6.6f, %6.6f, %6.6f)", P.x, P.y, P.z);
      Erro(mens_erro);
      Erro("\nERRO: no calculo da funcao DeterminaPontoDaIntersecao()");
      return 0;
    }

    P1.x = F1(p1.u, p1.v), P1.y = F2(p1.u, p1.v), P1.z = F3(p1.u, p1.v);
    P2.x = G1(q1.u, q1.v), P2.y = G2(q1.u, q1.v), P2.z = G3(q1.u, q1.v);
    if (!VetorNormal(&n1, F1, F2, F3, p1.u, p1.v)) {
      Erro("\n\nNao foi possivel calcular vetor normal a F");
      sprintf(mens_erro, "\nquando p = (%6.6f, %6.6f)", p1.u, p1.v);
      Erro(mens_erro);
      Erro("\nERRO: no calculo da funcao DeterminaPontoDaIntersecao()");
      return 0;
    }
    if (!VetorNormal(&n2, G1, G2, G3, q1.u, q1.v)) {
      Erro("\n\nNao foi possivel calcular vetor normal a G");
      sprintf(mens_erro, "\nquando q = (%6.6f, %6.6f)", q1.u, q1.v);
      Erro(mens_erro);
      Erro("\nERRO: no calculo da funcao DeterminaPontoDaIntersecao()");
      return 0;
    }

    PROD_VET(n3.x, n3.y, n3.z, n1.x, n1.y, n1.z, n2.x, n2.y, n2.z);
    P3.x = (P1.x + P2.x)/2, P3.y = (P1.y + P2.y)/2, P3.z = (P1.z + P2.z)/2;
    if (!ResolveSistema3x3(n1.x, n1.y, n1.z, n1.x*P1.x + n1.y*P1.y + n1.z*P1.z,
			   n2.x, n2.y, n2.z, n2.x*P2.x + n2.y*P2.y + n2.z*P2.z,
			   n3.x, n3.y, n3.z, n3.x*P3.x + n3.y*P3.y + n3.z*P3.z,
			   &sol1, &sol2, &sol3)) {
      Erro("\n\nNao foi possivel resolver sistema linear 3 x 3");
      Erro("\nERRO: no calculo da funcao DeterminaPontoDaIntersecao()");
      return 0;
    }

    P_anterior = P;
    P.x = sol1, P.y = sol2, P.z = sol3;

    p0 = p1, q0 = q1;
    num_it++;

  } while ((fabs(P_anterior.x - P.x) > z_erro ||
	   fabs(P_anterior.y - P.y) > z_erro ||
	   fabs(P_anterior.z - P.z) > z_erro) && num_it < MAX_ITERACOES);

  if (num_it >= MAX_ITERACOES) {
    Erro("\n\nIteracoes demais");
    Erro("\nERRO: no calculo da funcao DeterminaPontoDaIntersecao()");
    return 0;
  }

  *P0 = PontoNoDominio(P, F1, F2, F3, p1);
  if (P0->u == INF || P0->v == INF) {
    Erro("\n\nNao foi possivel encontrar ponto no dominio de F");
    Erro("\nERRO: no calculo da funcao DeterminaPontoDaIntersecao()");
    return 0;
  }

  *Q0 = PontoNoDominio(P, G1, G2, G3, q1);
  if (Q0->u == INF || Q0->v == INF) {
    Erro("\n\nNao foi possivel encontrar ponto no domino de G");
    Erro("\nERRO: no calculo da funcao DeterminaPontoDaIntersecao()");
    return 0;
  }

  return num_it;

}

/* ----------------------------------------------------------------------- */

int ConvPontoIntersecao(ponto *P, vetor *T,
		 ponto_dominio *pF, ponto_dominio *pG,
		 double (*f1)(double, double), double (*f2)(double, double),
		 double (*f3)(double, double), double (*g1)(double, double),
		 double (*g2)(double, double), double (*g3)(double, double)) {

  int iteracoes = 0;
  static ponto P1 = {0, 0, 0};
  static ponto P2 = {1, 1, 1};

  iteracoes = DeterminaPontoDaIntersecao(*P, pF, pG, f1, f2, f3, g1, g2, g3);

  if (iteracoes) {
    if (PontoNaIntersecao(f1, f2, f3, *pF, g1, g2, g3, *pG)) {
      P->x = f1(pF->u, pF->v), P->y = f2(pF->u, pF->v),
      P->z = f3(pF->u, pF->v);
      P1 = P2;
      P2 = *P;
    }
    else {
      Erro("\n\nPonto nao pertence `a intersecao de F e G");
      Erro("\nERRO: no calculo da funcao ConvPontoIntersecao()");
      return 0;
    }
  }
  else {
    Erro("\n\nNao foi possivel calcular ponto da intersecao");
    Erro("\nERRO: no calculo da funcao ConvPontoIntersecao()");
    return 0;
  }

  if (!VetorTangente(T, f1, f2, f3, *pF, g1, g2, g3, *pG)) {
    Erro("\n\nVetor tangente estimado como diferenca entre pontos anteriores");
    T->x = P2.x - P1.x, T->y = P2.y - P1.y, T->z = P2.z - P1.z;
  }

  return iteracoes;
}

/* ---------------------------------------------------------------------- */

int CaminhadaPassoCircular(int total, ponto_dominio p0, ponto_dominio q0,
	       ponto P, vetor v1, ponto Q, vetor v2, double L,
	       double (*f1)(double, double), double (*f2)(double, double),
	       double (*f3)(double, double), double (*g1)(double, double),
	       double (*g2)(double, double), double (*g3)(double, double),
	       int sentido_inverso) {

/* Executa caminha com passo circular de comprimento L. Sinaliza o final
   do tracado da curva no arquivo de dados padrao com:
	  FIM(1) ---> Curva chegou na fronteira do dominio
	  FIM(2) ---> Quantidade maxima de pontos atingida
	  FIM(3) ---> Problema de ma' aproximacao ou nao convergencia
	  FIM(4) ---> Retorno ao ponto inicial
   e o ponto inicial com
	  INICIO(1) ---> Caminhada no sentido do vetor tangente
	  INICIO(2) ---> Caminhada em sentido contrario ao vetor tangente
*/

  ponto R, R_copia, C;
  ponto_dominio pini, qini, p0_copia, q0_copia;
  int ponto_inicial = 1, quant = 0, iteracoes = 0;
  vetor T;
  double raio;
  static int vez = 1;

  if (vez == 1) {
    vez = 2;
    fprintf(arq_padrao, "(ponto inters)");
    if (ARQ_DOMINIO) fprintf(arq_padrao, " (pt dom F) (pt dom G)");
    if (ARQ_VETORES) fprintf(arq_padrao, " (vetor tg)");
    if (ARQ_CIRC)    fprintf(arq_padrao, " { (centro circ) raio }");
    if (ITERACOES)   fprintf(arq_padrao, " [iter]");
  }

  do {

    if (!ProximoPontoCirc(&R, &C, &raio, P, v1, Q, v2, L)) {
      sprintf(mens_erro,"\n\nP = (%6.6f, %6.6f, %6.6f)  v1 = (%6.6f, %6.6f,"
				 " %6.6f)", P.x, P.y, P.z, v1.x, v1.y, v1.z);
      Erro(mens_erro);
      sprintf(mens_erro,"\nQ = (%6.6f, %6.6f, %6.6f)  v2 = (%6.6f, %6.6f,"
				 " %6.6f)", Q.x, Q.y, Q.z, v2.x, v2.y, v2.z);
      Erro(mens_erro);
      sprintf(mens_erro,"\nL = %6.6f", L);
      Erro(mens_erro);
      Erro("\nMudando temporariamente para passo tangencial.");
      R = Q;
      ProximoPontoVetorTangente(&R, v2, L); /* Muda para passo tangencial */
      /* return 0; */
    }

    P = Q, v1 = v2;

    R_copia = R, p0_copia = p0, q0_copia = q0;

    iteracoes = ConvPontoIntersecao(&R, &T, &p0, &q0,
					f1, f2, f3, g1, g2, g3);
    if (iteracoes) {
      if (DOMINIO_LIMITADO)
	if (p0.u < Ret1.xmin || p0.u > Ret1.xmax ||
	   p0.v < Ret1.ymin || p0.v > Ret1.ymax ||
	   q0.u < Ret2.xmin || q0.u > Ret2.xmax ||
	   q0.v < Ret2.ymin || q0.v > Ret2.ymax) {
	     Erro("\n\nCurva chegou na fronteira do dominio.");
	     fprintf(arq_padrao, " FIM(1)");
	     return 0;
	   }
    } else {
      sprintf(mens_erro,"\n\nR = (%6.6f, %6.6f, %6.6f)  p0 = (%6.6f, %6.6f)  "
      "q0 = (%6.6f, %6.6f)", R_copia.x, R_copia.y, R_copia.z, p0_copia.u,
					p0_copia.v, q0_copia.u, q0_copia.v);
      Erro(mens_erro);
      Erro("\nERRO: no calculo da funcao ConvPontoIntersecao()");
      fprintf(arq_padrao, " FIM(3)");
      return 0;
    }

    Q = R, v2 = T;

    fprintf(arq_padrao, "\n(%0.*f %0.*f %0.*f)",
			 ARQ_DEC, R.x, ARQ_DEC, R.y, ARQ_DEC, R.z);
    if (ARQ_DOMINIO) {
      fprintf(arq_padrao, " (%0.*f %0.*f)",
			 ARQ_DEC, p0.u, ARQ_DEC, p0.v);
      fprintf(arq_padrao, " (%0.*f %0.*f)",
			 ARQ_DEC, q0.u, ARQ_DEC, q0.v);
    }
    if (ARQ_VETORES)
      fprintf(arq_padrao, " (%0.*f %0.*f %0.*f)",
			 ARQ_DEC, T.x, ARQ_DEC, T.y, ARQ_DEC, T.z);
    if (ARQ_CIRC)
      fprintf(arq_padrao, " {(%0.*f %0.*f %0.*f) %0.*f}",
	    ARQ_DEC, C.x, ARQ_DEC, C.y, ARQ_DEC, C.z, ARQ_DEC, raio);
    if (ITERACOES)
      fprintf(arq_padrao, " [%d]", iteracoes);

    if (ponto_inicial) {
      fprintf(arq_padrao, (sentido_inverso? " INICIO(2)" : " INICIO(1)"));
      ponto_inicial = 0;
      pini = p0, qini = q0;
    } else {
      if (VERIFICA_RETORNO) {
	if (DIST1(pini.u,pini.v,qini.u,qini.v,p0.u,p0.v,q0.u,q0.v)<r_erro) {
	  Erro("\nRetorno ao ponto inicial");
	  fprintf(arq_padrao, " FIM(4)");
	  return 0;
	}
      }
    }

  } while (++quant < total);

  fprintf(arq_padrao, " FIM(2)");
  return quant;
}

/* ---------------------------------------------------------------------- */

int CaminhadaPassoTangencial(int total, ponto_dominio p0, ponto_dominio q0,
	       ponto R, vetor T, double L,
	       double (*f1)(double, double), double (*f2)(double, double),
	       double (*f3)(double, double), double (*g1)(double, double),
	       double (*g2)(double, double), double (*g3)(double, double)) {

  int quant = 0, iteracoes = 0;
  ponto R_copia;
  ponto_dominio p0_copia, q0_copia;
  static int vez = 1;

  if (vez == 1) {
    vez = 2;
    fprintf(arq_padrao, "(ponto inters)");
    if (ARQ_DOMINIO) fprintf(arq_padrao, " (pt dom F) (pt dom G)");
    if (ARQ_VETORES) fprintf(arq_padrao, " (vetor tg)");
    if (ITERACOES)   fprintf(arq_padrao, " [iter]");
  }

  do {

    ProximoPontoVetorTangente(&R, T, L);

    R_copia = R, p0_copia = p0, q0_copia = q0;

    iteracoes = ConvPontoIntersecao(&R, &T, &p0, &q0,
					f1, f2, f3, g1, g2, g3);
    if (iteracoes) {
      if (DOMINIO_LIMITADO)
	if (p0.u < Ret1.xmin || p0.u > Ret1.xmax ||
	  p0.v < Ret1.ymin || p0.v > Ret1.ymax ||
	  q0.u < Ret2.xmin || q0.u > Ret2.xmax ||
	  q0.v < Ret2.ymin || q0.v > Ret2.ymax) {
	    Erro("\n\nCurva chegou na fronteira do dominio.");
	    return 0;
	  }
    } else {
      sprintf(mens_erro,"\n\nR = (%6.6f, %6.6f, %6.6f)  p0 = (%6.6f, %6.6f)  "
      "q0 = (%6.6f, %6.6f)", R_copia.x, R_copia.y, R_copia.z, p0_copia.u,
					p0_copia.v, q0_copia.u, q0_copia.v);
      Erro(mens_erro);
      Erro("\nERRO: no calculo da funcao ConvPontoIntersecao()");
      return 0;
    }

    fprintf(arq_padrao, "\n(%0.*f %0.*f %0.*f)",
			 ARQ_DEC, R.x, ARQ_DEC, R.y, ARQ_DEC, R.z);
    if (ARQ_DOMINIO) {
      fprintf(arq_padrao, " (%0.*f %0.*f)",
			 ARQ_DEC, p0.u, ARQ_DEC, p0.v);
      fprintf(arq_padrao, " (%0.*f %0.*f)",
			 ARQ_DEC, q0.u, ARQ_DEC, q0.v);
    }
    if (ARQ_VETORES)
      fprintf(arq_padrao, " (%0.*f %0.*f %0.*f)",
			 ARQ_DEC, T.x, ARQ_DEC, T.y, ARQ_DEC, T.z);
    if (ITERACOES)
      fprintf(arq_padrao, " [%d]", iteracoes);

  } while (++quant < total);
  return quant;
}

/* ----------------------------------------------------------------------- */

unsigned long CalculaIteracoes(char *NomeArq) {

  FILE *arq;
  unsigned long total_pontos = 0, total_it = 0;
  char *ch, linha[200];
  float x, y, z, t1, t2, t3, u1, v1, u2, v2;
  unsigned long it0 = 0, it1 = 0, it2 = 0, it3 = 0, it4 = 0,
		it5 = 0, it6 = 0, it7 = 0, it8 = 0, it9 = 0, it = 0;

  arq = fopen(NomeArq, "rb");
  if (arq == NULL) return 0;

  fgets(linha, 200, arq);

  do {
    if ((ch = fgets(linha, 200, arq)) != NULL) {
      sscanf(linha, "(%f %f %f) (%f %f) (%f %f) (%f %f %f) [%d]",
		      &x, &y, &z, &u1, &v1, &u2, &v2, &t1, &t2, &t3, &it);

      total_pontos++;
      if (it == 0) it0++;
      else if (it == 1) it1++;
      else if (it == 2) it2++;
      else if (it == 3) it3++;
      else if (it == 4) it4++;
      else if (it == 5) it5++;
      else if (it == 6) it6++;
      else if (it == 7) it7++;
      else if (it == 8) it8++;
      else if (it >= 9) it9++;

      total_it += it;
    }
  } while (ch != NULL);

  if (MOSTRA_HIST) {
    if (it0) printf("\n       Pontos que convergem com 0 iteracao: %lu (%2.2f %c)",
		      it0, (float) (it0*100.0/total_pontos), 37);
    if (it1) printf("\n       Pontos que convergem com 1 iteracao: %lu (%2.2f %c)",
		      it1, (float) (it1*100.0/total_pontos), 37);
    if (it2) printf("\n       Pontos que convergem com 2 iteracoes: %lu (%2.2f %c)",
		      it2, (float) (it2*100.0/total_pontos), 37);
    if (it3) printf("\n       Pontos que convergem com 3 iteracoes: %lu (%2.2f %c)",
		      it3, (float) (it3*100.0/total_pontos), 37);
    if (it4) printf("\n       Pontos que convergem com 4 iteracoes: %lu (%2.2f %c)",
		      it4, (float) (it4*100.0/total_pontos), 37);
    if (it5) printf("\n       Pontos que convergem com 5 iteracoes: %lu (%2.2f %c)",
		      it5, (float) (it5*100.0/total_pontos), 37);
    if (it6) printf("\n       Pontos que convergem com 6 iteracoes: %lu (%2.2f %c)",
		      it6, (float) (it6*100.0/total_pontos), 37);
    if (it7) printf("\n       Pontos que convergem com 7 iteracoes: %lu (%2.2f %c)",
		      it7, (float) (it7*100.0/total_pontos), 37);
    if (it8) printf("\n       Pontos que convergem com 8 iteracoes: %lu (%2.2f %c)",
		      it8, (float) (it8*100.0/total_pontos), 37);
    if (it9) printf("\n       Pontos que convergem com +9 iteracoes: %lu (%2.2f %c)",
		      it9, (float) (it9*100.0/total_pontos), 37);
  }
  if (SALVA_HIST) {
    if (it0) fprintf(arq_execucao, "\n       Pontos que convergem com 0 iteracao: %lu (%2.2f %c)",
		      it0, (float) (it0*100.0/total_pontos), 37);
    if (it1) fprintf(arq_execucao, "\n       Pontos que convergem com 1 iteracao: %lu (%2.2f %c)",
		      it1, (float) (it1*100.0/total_pontos), 37);
    if (it2) fprintf(arq_execucao, "\n       Pontos que convergem com 2 iteracoes: %lu (%2.2f %c)",
		      it2, (float) (it2*100.0/total_pontos), 37);
    if (it3) fprintf(arq_execucao, "\n       Pontos que convergem com 3 iteracoes: %lu (%2.2f %c)",
		      it3, (float) (it3*100.0/total_pontos), 37);
    if (it4) fprintf(arq_execucao, "\n       Pontos que convergem com 4 iteracoes: %lu (%2.2f %c)",
		      it4, (float) (it4*100.0/total_pontos), 37);
    if (it5) fprintf(arq_execucao, "\n       Pontos que convergem com 5 iteracoes: %lu (%2.2f %c)",
		      it5, (float) (it5*100.0/total_pontos), 37);
    if (it6) fprintf(arq_execucao, "\n       Pontos que convergem com 6 iteracoes: %lu (%2.2f %c)",
		      it6, (float) (it6*100.0/total_pontos), 37);
    if (it7) fprintf(arq_execucao, "\n       Pontos que convergem com 7 iteracoes: %lu (%2.2f %c)",
		      it7, (float) (it7*100.0/total_pontos), 37);
    if (it8) fprintf(arq_execucao, "\n       Pontos que convergem com 8 iteracoes: %lu (%2.2f %c)",
		      it8, (float) (it8*100.0/total_pontos), 37);
    if (it9) fprintf(arq_execucao, "\n       Pontos que convergem com +9 iteracoes: %lu (%2.2f %c)",
		      it9, (float) (it9*100.0/total_pontos), 37);
  }
  fclose(arq);

  return total_pontos;
}

/* ----------------------------------------------------------------------- */

int Simplifica(char *ArqOrigem, char *ArqDestino, float erro, int opcao) {

  FILE *arq, *arq2;
  unsigned long total = 0, quant = 0;
  char *ch, linha[200];
  float x, y, z, t1, t2, t3, u1, v1, u2, v2;
  unsigned long i, novo = 1;
  ponto_R4 *p;

  if ((arq = fopen(ArqOrigem, "rb")) == NULL) return 0;
  if ((arq2 = fopen(ArqDestino, "wb")) == NULL) return 0;

  if ((p = (ponto_R4 *) malloc((MAX_DISTINTOS+1)*sizeof(ponto_R4))) == NULL)
    return 0;

  fgets(linha, 200, arq);
  fprintf(arq2, linha);

  do {
    ch = fgets(linha, 200, arq);
    if (ch != NULL) {
      sscanf(linha, "(%f %f %f) (%f %f) (%f %f) (%f %f %f)",
		      &x, &y, &z, &u1, &v1, &u2, &v2, &t1, &t2, &t3);
	total++;
	novo = 1;
	for (i = 0; i < quant; i++)
	  if (opcao == 1) {
	    if (DIST1(u1, v1, u2, v2, p[i].x, p[i].y, p[i].z, p[i].t) < erro)
	      { novo = 0; break; }
	  } else if (opcao == 2) {
	    if (DIST2(u1, v1, u2, v2, p[i].x, p[i].y, p[i].z, p[i].t) < erro)
	      { novo = 0; break; }
	  } else if (opcao == 3) {
	    if (DIST3(u1, v1, p[i].x, p[i].y) < erro &&
		DIST3(u2, v2, p[i].z, p[i].t) < erro) { novo = 0; break; }
	  }

	if (novo) {
	  fprintf(arq2, linha);
	  p[quant].x = u1, p[quant].y = v1, p[quant].z = u2, p[quant].t = v2;
	  if (quant < MAX_DISTINTOS) quant++; else goto fim;
	}
    }
  } while (ch != NULL);

  fim: fclose(arq); fclose(arq2);
  if (MOSTRA_HIST) {
    printf("\n\n       Total de pontos distintos: %lu (%4.2f %c)",
				       quant, (quant*100.0/total), 37);
    printf("\n       Distancia minima entre os pontos: %g", erro);
  }
  if (SALVA_HIST) {
    fprintf(arq_execucao, "\n\n       Total de pontos distintos: %lu (%4.2f %c)",
				       quant, (quant*100.0/total), 37);
    fprintf(arq_execucao, "\n       Distancia minima entre os pontos: %g", erro);
  }
  return 1;
}

/* ----------------------------------------------------------------------- */

/*** FIM DE CONVERG.C ***/
