/*
     []-------------------------------------------------------------[]
     |		                SUBDIV.C                             |
     |						 	             |
     |     Metodo da subdivisao para a determinacao da intersecao    |
     |                    entre duas superficies                     |
     |						                     |
     |                por Lenimar N. Andrade, maio/97                |
     []-------------------------------------------------------------[]
*/

/* Quantidades de subdivisoes utilizadas em cada nivel */
int N_SUBDIV1 = 80, N_SUBDIV2 = 40, N_SUBDIV3 = 20, N_SUBDIV4 = 10,
    N_SUBDIV5 = 10;

/* int NIVEL_CLASS = 1, DOMINIO2 = 0; */

int DIV_RASTREAR,       /* Define se mostra valores encontrados na tela */
    DIV_ARQ_DOMINIO,    /* Define se salva pontos do dominio            */
    DIV_ARQ_VALORES;    /* Define se salva valores de pontos do dominio */

char DIV_NOME_ARQ[30] = "subdiv.txt", /* Nome do arquivo de dados       */
DIV_CLASS1[30] = "class1.txt",        /* Algumas linhas de DIV_NOME_ARQ */
DIV_CLASS2[30] = "class2.txt";        /* Algumas linhas de DIV_NOME_ARQ */

/* Verifica se o ponto (a, b, c) pertence ao interior da caixa C */
#define INTERIOR(a, b, c, C)  (C.x<=a&&a<=C.X&&C.y<=b&&b<=C.Y&&C.z<=c&&c<=C.Z)

/* Variaveis globais */
retangulo DIV_Ret1_copia, DIV_Ret2_copia;
FILE *DIV_arq_dados;
char DIV_linha_dist_min[100];

#ifndef UNIX
extern unsigned _stklen = 20000;     /* Ajusta tamanho da pilha */
#endif

#define COORD_CARTESIANAS     /* Salva coord. cartesianas no arq. subdiv. */

/* ---------------------------------------------------------------------- */

void DefineCaixa(caixa *C, int N, retangulo R, double (*f1)(double, double),
	       double (*f2)(double, double), double (*f3)(double, double)) {

/* Define uma caixa ('bounding box') que limite F(u, v) quando (u, v) for
   restrito ao dominio R. Sao calculados maximos e minimos de F. */

  double x, y, incrX, incrY;

  incrX = (R.xmax - R.xmin)/N;
  incrY = (R.ymax - R.ymin)/N;
  R.xmax += 0.5*incrX;
  R.ymax += 0.5*incrY;

  C->X = C->Y = C->Z = -INF;
  C->x = C->y = C->z = INF;
  for (x = R.xmin; x <= R.xmax; x += incrX)
    for (y = R.ymin; y <= R.ymax; y += incrY) {
      if (f1(x, y) > C->X) C->X = f1(x, y);
      if (f1(x, y) < C->x) C->x = f1(x, y);
      if (f2(x, y) > C->Y) C->Y = f2(x, y);
      if (f2(x, y) < C->y) C->y = f2(x, y);
      if (f3(x, y) > C->Z) C->Z = f3(x, y);
      if (f3(x, y) < C->z) C->z = f3(x, y);
    }
}

/* ---------------------------------------------------------------------- */

int IntCaixas(caixa A, caixa B) {

/* Verifica se as caixas A e B se interceptam. Retorna 1 se houver
   intersecao ou 0 em caso contrario. */

  if (INTERIOR(A.x, A.y, A.z, B)) return 1;
  if (INTERIOR(A.X, A.y, A.z, B)) return 1;
  if (INTERIOR(A.x, A.Y, A.z, B)) return 1;
  if (INTERIOR(A.X, A.Y, A.z, B)) return 1;
  if (INTERIOR(A.x, A.y, A.Z, B)) return 1;
  if (INTERIOR(A.X, A.y, A.Z, B)) return 1;
  if (INTERIOR(A.x, A.Y, A.Z, B)) return 1;
  if (INTERIOR(A.X, A.Y, A.Z, B)) return 1;

  if (INTERIOR(B.x, B.y, B.z, A)) return 1;
  if (INTERIOR(B.X, B.y, B.z, A)) return 1;
  if (INTERIOR(B.x, B.Y, B.z, A)) return 1;
  if (INTERIOR(B.X, B.Y, B.z, A)) return 1;
  if (INTERIOR(B.x, B.y, B.Z, A)) return 1;
  if (INTERIOR(B.X, B.y, B.Z, A)) return 1;
  if (INTERIOR(B.x, B.Y, B.Z, A)) return 1;
  if (INTERIOR(B.X, B.Y, B.Z, A)) return 1;

  if (A.x <= B.x && B.X <= A.X && A.y <= B.y && B.Y <= A.Y &&
      B.z <= A.z && A.Z <= B.Z) return 1;
  if (B.x <= A.x && A.X <= B.X && B.y <= A.y && A.Y <= B.Y &&
      A.z <= B.z && B.Z <= A.Z) return 1;

  return 0;
}

/* ---------------------------------------------------------------------- */

retangulo SelecionaRetangulo(int n, retangulo X) {

/*  Seleciona um subretangulo 1, 2, 3 ou 4 do retangulo X.

	       X.ymax -- +-----+-----+
			 |  1  |  2  |
			 +-----+-----+
			 |  3  |  4  |
	       X.ymin -- +-----+-----+
			 |           |
		      X.xmin      X.xmax
*/

  retangulo R;

  if (n == 1) { R.xmin = X.xmin, R.xmax = (X.xmin + X.xmax)/2,
		R.ymin = (X.ymin + X.ymax)/2, R.ymax = X.ymax; }

  else if (n == 2) { R.xmin = (X.xmin + X.xmax)/2, R.xmax = X.xmax,
		R.ymin = (X.ymin + X.ymax)/2, R.ymax = X.ymax; }

  else if (n == 3) { R.xmin = X.xmin, R.xmax = (X.xmin + X.xmax)/2,
		R.ymin = X.ymin, R.ymax = (X.ymin + X.ymax)/2; }

  else { R.xmin = (X.xmin + X.xmax)/2, R.xmax = X.xmax,
		R.ymin = X.ymin, R.ymax = (X.ymin + X.ymax)/2; }

  return R;
}

/* ---------------------------------------------------------------------- */

void InterceptaCaixas(int *r, int *s, retangulo A, retangulo B, int n_subdiv,
	       double (*f1)(double, double), double (*f2)(double, double),
	       double (*f3)(double, double), double (*g1)(double, double),
	       double (*g2)(double, double), double (*g3)(double, double)) {

/* Determina os pares de caixas  que se  interceptam.  Os  pares que se
   interceptam sao (r[0], s[0]), (r[1], s[1]), etc. Sao usadas n_subdiv
   subdivisoes   dos   intervalos  [A.xmin, A.xmax],  [A.ymin, A.ymax],
   [B.xmin, B.xmax], [B.ymin, B.ymax]. */

  caixa X[5], Y[5];
  retangulo R, S;
  int cont = 0, i, j;

  r[cont] = 0, s[cont] = 0;

  for (i = 1; i <= 4; i++) {
    R = SelecionaRetangulo(i, A);
    S = SelecionaRetangulo(i, B);
    DefineCaixa(&X[i], n_subdiv, R, f1, f2, f3);
    DefineCaixa(&Y[i], n_subdiv, S, g1, g2, g3);
  }

  for (i = 1; i <= 4; i++)
    for (j = 1; j <= 4; j++) {
      if (IntCaixas(X[i], Y[j])) {
	r[cont] = i, s[cont] = j;
	cont++;
	r[cont] = 0, s[cont] = 0;
      }
    }
}

/* ---------------------------------------------------------------------- */

void PontoEmRetangulo(char *r, retangulo R, double *x, double *y) {

/* Determina que ponto do dominio esta' associado `a representacao r */

  int i = 0;

  while (r[i]) {
    if (r[i] == '1') R = SelecionaRetangulo(1, R);
    else if (r[i] == '2') R = SelecionaRetangulo(2, R);
    else if (r[i] == '3') R = SelecionaRetangulo(3, R);
    else R = SelecionaRetangulo(4, R);
    i++;
  }
  *x = (R.xmin + R.xmax)/2, *y = (R.ymin + R.ymax)/2;
}

/* ---------------------------------------------------------------------- */

void RetanguloQueContemPonto(double x, double y, retangulo R, char *r,
								 int nivel) {

/* Determina a representacao r de um ponto (x, y) do dominio. Se nivel = 6,
   a representacao de um ponto pode ser algo do tipo "112143" (por ex.) */

  int i, j;

  if (x < R.xmin || x > R.xmax || y < R.ymin || y > R.ymax) {
    r[0] = '\0';
    return;
  }

  for (i = 1; i <= nivel; i++) {
    if (x < (R.xmin + R.xmax)/2) {
      if (y < (R.ymin + R.ymax)/2)
	{ r[i - 1] = '3', j = 3; }
      else
	{ r[i - 1] = '1'; j = 1; }
    } else {
      if (y < (R.ymin + R.ymax)/2)
	{ r[i - 1] = '4'; j = 4; }
      else
	{ r[i - 1] = '2'; j = 2; }
    }
    r[i] = '\0';
    R = SelecionaRetangulo(j, R);
  }
}

/* ---------------------------------------------------------------------- */

int LePontoSubdivDominio(char *nome_arq, retangulo R1, retangulo R2,
	  ponto_dominio *p, ponto_dominio *q, char *linha, long n) {

/* Retorna o n-esimo ponto do arquivo nome_arq */

  FILE *arq;
  char *ch, linha1[100], linha2[100];
  float distancia;
  long quant = 0;

  if ((arq = fopen(nome_arq, "rb")) == NULL) {
    sprintf(mens_erro, "\n\nArquivo %s nao pode ser aberto", nome_arq);
    Erro(mens_erro);
    Erro("\nERRO: no calculo da funcao LePontoSubdivDominio()");
    return 0;
  }

  do {
    ch = fgets(linha, 150, arq);
    if (ch != NULL)
      if (linha[0] == '1' || linha[0] == '2' || linha[0] == '3' ||
						linha[0] == '4') {
	quant++;
	if (quant == n) {
	  sscanf(linha, "%s %s [%f]", linha1, linha2, &distancia);
	  PontoEmRetangulo(linha1, R1, &p->u, &p->v);
	  PontoEmRetangulo(linha2, R2, &q->u, &q->v);
	  fclose(arq);
	  return 1;
	}
      }
  } while (ch != NULL);

  fclose(arq);
  return 0;
}

/* ---------------------------------------------------------------------- */

long ContaPontoSubdivDominio(char *nome_arq) {

/* Conta quantos pontos existem no arquivo nome_arq */

  FILE *arq;
  char *ch, linha[100];
  long quant = 0;

  if ((arq = fopen(nome_arq, "rb")) == NULL) {
    sprintf(mens_erro, "\n\nArquivo %s nao pode ser aberto", nome_arq);
    Erro(mens_erro);
    Erro("\nERRO: no calculo da funcao ContaPontoSubdivDominio()");
    return 0;
  }

  do {
    ch = fgets(linha, 100, arq);
    if (ch != NULL)
      if (linha[0] == '1' || linha[0] == '2' || linha[0] == '3' ||
						linha[0] == '4') quant++;
  } while (ch != NULL);

  fclose(arq);
  return quant;
}

/* ---------------------------------------------------------------------- */

int LeOutroPonto(ponto_dominio *p, ponto_dominio *q, char *linha, int opcao,
								int *novo) {

/* Retorna outro ponto obtido na subdivisao que tenha o mesmo padrao da
   linha apresentada. */

  FILE *arq;
  char padrao[100], nome_arq[100], *ch, linha1[100], linha2[100], x[100];
  float distancia;
  long quant = 0;
  static long n_ponto_atual = 0;
  static long n_pontos_arq = 0;

  sscanf(linha, "%s %s [%f]", linha1, linha2, &distancia);
  if (opcao == 1) {
    strcpy(padrao, linha1);
    padrao[NIVEL_CLASS] = '\0';
    sprintf(nome_arq, "_%s.tmp", padrao);
  } else {
    strcpy(padrao, linha2);
    padrao[NIVEL_CLASS] = '\0';
    sprintf(nome_arq, "~%s.tmp", padrao);
  }

  if (*novo) {
    n_ponto_atual = 0;
    n_pontos_arq = ContaPontoSubdivDominio(nome_arq);
  }
  n_ponto_atual++;

  if (n_ponto_atual > n_pontos_arq) { *novo = 1; return 0; } else *novo = 0;

  if ((arq = fopen(nome_arq, "rb")) == NULL) {
    sprintf(mens_erro, "\n\nArquivo %s nao pode ser aberto", nome_arq);
    Erro(mens_erro);
    Erro("\nERRO: no calculo da funcao LeOutroPonto()");
    return 0;
  }

  do {
    ch = fgets(x, 100, arq);
    if (ch != NULL)
      if (x[0] == '1' || x[0] == '2' || x[0] == '3' || x[0] == '4') {
	quant++;
	if (quant == n_ponto_atual) {
	  sscanf(x, "%s %s [%f]", linha1, linha2, &distancia);
	  PontoEmRetangulo(linha1, Ret1, &p->u, &p->v);
	  PontoEmRetangulo(linha2, Ret2, &q->u, &q->v);
	  fclose(arq);
	  return 1;
	}
      }
  } while (ch != NULL);

  fclose(arq);
  *novo = 1;
  return 0;
}

/* ---------------------------------------------------------------------- */

long ProcuraLinhaPadrao(char *nome_arq, char *padrao, int opcao) {

/* Procura linhas em  arquivo que  combinam com  determinado padrao.
   Interpreta os valores em linha do tipo
	      "representacao1 representacao2 [distancia]"
   Salva na variavel global DIV_linha_dist_min a linha que estiver asso-
   ciada a uma distancia minima. Salva linhas em um arquivo auxiliar.
   Se opcao == 1 a pesquisa e' feita no dominio de F; se opcao == 2, e'
   feita no dominio de G. */

  FILE *arq;
  char *ch, linha[100], linha1[100], linha2[100];
  long quant = 0;
  float dmin = INF, distancia;
  int tam_padrao;

  if (opcao != 1 && opcao != 2) opcao = 1;
  tam_padrao = strlen(padrao);

  arq = fopen(nome_arq, "rb");
  if (arq == NULL) return -1;

  do {
    ch = fgets(linha, 100, arq);
    if (ch != NULL)
      if (linha[0] == '1' || linha[0] == '2' || linha[0] == '3' ||
						linha[0] == '4') {
	sscanf(linha, "%s %s [%f]", linha1, linha2, &distancia);
	if ((opcao == 1 && !strncmp(linha1, padrao, tam_padrao)) ||
	    (opcao == 2 && !strncmp(linha2, padrao, tam_padrao))) {
	  quant++;
	  if (dmin > distancia) strcpy(DIV_linha_dist_min, linha);
	}
      }
  } while (ch != NULL);

  fclose(arq);
  return quant;
}

/* ---------------------------------------------------------------------- */

int FragmentaArquivoSubdiv(char *nome_arq, int nivel) {

/* Procura linhas em  arquivo que  combinam com  determinado padrao.
   Interpreta os valores em linha do tipo
	      "representacao1 representacao2 [distancia]"    */

  FILE *arq, *arq2;
  char *ch, linha[100], linha1[100], linha2[100], nome_arq2[100];
  float distancia;

  arq = fopen(nome_arq, "rb");
  if (arq == NULL) {
    sprintf(mens_erro, "\n\nArquivo %s nao pode ser aberto", nome_arq);
    Erro(mens_erro);
    Erro("\nERRO: no calculo da funcao FragmentaArquivoSubdiv()");
    return 0;
  }

  do {
    ch = fgets(linha, 100, arq);
    if (ch != NULL)
      if (linha[0] == '1' || linha[0] == '2' || linha[0] == '3' ||
						linha[0] == '4') {
	sscanf(linha, "%s %s [%f]", linha1, linha2, &distancia);
	linha1[nivel] = linha2[nivel] = '\0';

	sprintf(nome_arq2, "_%s.tmp", linha1);
	arq2 = fopen(nome_arq2, "ab");
	if (arq2 == NULL) {
	  sprintf(mens_erro, "\n\nArquivo %s nao pode ser aberto", nome_arq2);
	  Erro(mens_erro);
	  Erro("\nERRO: no calculo da funcao FragmentaArquivoSubdiv()");
	  return 0;
	}
	fprintf(arq2, linha);
	fclose(arq2);
	if (DOMINIO2) {
	  sprintf(nome_arq2, "~%s.tmp", linha2);
	  arq2 = fopen(nome_arq2, "ab");
	  if (arq2 == NULL) {
	    sprintf(mens_erro, "\n\nArquivo %s nao pode ser aberto", nome_arq2);
	    Erro(mens_erro);
	    Erro("\nERRO: no calculo da funcao FragmentaArquivoSubdiv()");
	    return 0;
	  }
	  fprintf(arq2, linha);
	  fclose(arq2);
	}
      }
  } while (ch != NULL);

  fclose(arq);
  return 1;
}

/* ---------------------------------------------------------------------- */

int ProcuraLinhaDist(char *nome_arq, double dist_max) {

/* Procura linhas em arquivo com distancias menores ou iguais a dist_max.
   Se dist_max for negativo,  entao e'  encontrada  a linha de distancia
   minima. */

  FILE *arq;
  char *ch, linha_aux[100], linha1[100], linha2[100];
  float dist, dist_min = INF;

  arq = fopen(nome_arq, "rb");
  if (arq == NULL) {
    sprintf(mens_erro, "\n\nArquivo %s nao pode ser aberto", nome_arq);
    Erro(mens_erro);
    Erro("\nERRO: no calculo da funcao ProcuraLinhasDist()");
    return 0;
  }

  do {
    ch = fgets(linha_aux, 100, arq);
    if (ch != NULL)
      if (linha_aux[0] == '1' || linha_aux[0] == '2' ||
	  linha_aux[0] == '3' || linha_aux[0] == '4') {
	sscanf(linha_aux, "%s %s [%f]", linha1, linha2, &dist);

	if (dist_max >= 0) {
	  if (dist <= dist_max) {
	    strcpy(DIV_linha_dist_min, linha_aux);
	    fclose(arq);
	    return 1;
	   }
	} else {
	  if (dist_min > dist) {
	    dist_min = dist;
	    strcpy(DIV_linha_dist_min, linha_aux);
	  }
	}
      }
  } while (ch != NULL);

  fclose(arq);
  if (dist_max < 0) return 1;
  return 0;
}

/* ---------------------------------------------------------------------- */

char i2c(int x) {

   if (x == 1) return '1';
   else if (x == 2) return '2';
   else if (x == 3) return '3';
   else if (x == 4) return '4';
   return '\0';
}

/* ---------------------------------------------------------------------- */

long SelecionaPontos(char *nome_arq, int nivel, int opcao) {

/* Seleciona pontos do arquivo nome_arq, de modo a ficarem bem distribuidos
   ao longo do dominio. Se opcao == 1, e' usado o dominio de F como referen-
   cia; se opcao == 2, e' usado o dominio de G. */

  long quant = 0;
  FILE *arq, *arq2;
  int i, j, k, w, t;
  char padrao[100], nome_arq2[100];

  arq = fopen(nome_arq, "wb");
  if (arq == NULL) {
    sprintf(mens_erro, "\n\nArquivo %s nao pode ser aberto", nome_arq);
    Erro(mens_erro);
    Erro("\nERRO: no calculo da funcao SelecionaPontos()");
    return 0;
  }

  if (nivel < 1) nivel = 1; else if (nivel > 5) nivel = 5;

  if (nivel == 1) {
    for (i = 1; i <= 4; i++) {
      padrao[0] = i2c(i);
      padrao[1] = '\0';
      if (opcao == 1)
	sprintf(nome_arq2, "_%s.tmp", padrao);
      else
	sprintf(nome_arq2, "~%s.tmp", padrao);
      arq2 = fopen(nome_arq2, "rb");
      if (arq2 != NULL) {
	if (ProcuraLinhaDist(nome_arq2, -1)) {
	  quant++;
	  fprintf(arq, DIV_linha_dist_min);
	}
	fclose(arq2);
	/* remove(nome_arq2); */
      }
      else if (ProcuraLinhaPadrao(DIV_NOME_ARQ, padrao, opcao)) {
	quant++;
	fprintf(arq, DIV_linha_dist_min);
      }
    }

  } else if (nivel == 2) {
    for (i = 1; i <= 4; i++)
      for (j = 1; j <= 4; j++) {
	padrao[0] = i2c(i);
	padrao[1] = i2c(j);
	padrao[2] = '\0';
	if (opcao == 1)
	  sprintf(nome_arq2, "_%s.tmp", padrao);
	else
	  sprintf(nome_arq2, "~%s.tmp", padrao);
	arq2 = fopen(nome_arq2, "rb");
	if (arq2 != NULL) {
	  if (ProcuraLinhaDist(nome_arq2, -1)) {
	    quant++;
	    fprintf(arq, DIV_linha_dist_min);
	  }
	  fclose(arq2);
	}
	else if (ProcuraLinhaPadrao(DIV_NOME_ARQ, padrao, opcao)) {
	  quant++;
	  fprintf(arq, DIV_linha_dist_min);
	}
      }

  } else if (nivel == 3) {
    for (i = 1; i <= 4; i++)
      for (j = 1; j <= 4; j++)
	for (k = 1; k <= 4; k++) {
	  padrao[0] = i2c(i);
	  padrao[1] = i2c(j);
	  padrao[2] = i2c(k);
	  padrao[3] = '\0';
	  if (opcao == 1)
	    sprintf(nome_arq2, "_%s.tmp", padrao);
	  else
	    sprintf(nome_arq2, "~%s.tmp", padrao);
	  arq2 = fopen(nome_arq2, "rb");
	  if (arq2 != NULL) {
	    if (ProcuraLinhaDist(nome_arq2, -1)) {
	      quant++;
	      fprintf(arq, DIV_linha_dist_min);
	    }
	    fclose(arq2);
	  }
	  else if (ProcuraLinhaPadrao(DIV_NOME_ARQ, padrao, opcao)) {
	    quant++;
	    fprintf(arq, DIV_linha_dist_min);
	  }
	}

  } else if (nivel == 4) {
    for (i = 1; i <= 4; i++)
      for (j = 1; j <= 4; j++)
	for (k = 1; k <= 4; k++)
	  for (w = 1; w <= 4; w++) {
	    padrao[0] = i2c(i);
	    padrao[1] = i2c(j);
	    padrao[2] = i2c(k);
	    padrao[3] = i2c(w);
	    padrao[4] = '\0';
	    if (opcao == 1)
	      sprintf(nome_arq2, "_%s.tmp", padrao);
	    else
	      sprintf(nome_arq2, "~%s.tmp", padrao);
	    arq2 = fopen(nome_arq2, "rb");
	    if (arq2 != NULL) {
	      if (ProcuraLinhaDist(nome_arq2, -1)) {
		quant++;
		fprintf(arq, DIV_linha_dist_min);
	      }
	      fclose(arq2);
	    }
	    else if (ProcuraLinhaPadrao(DIV_NOME_ARQ, padrao, opcao)) {
	      quant++;
	      fprintf(arq, DIV_linha_dist_min);
	    }
	}

  } else if (nivel == 5) {
    for (i = 1; i <= 4; i++)
      for (j = 1; j <= 4; j++)
	for (k = 1; k <= 4; k++)
	  for (w = 1; w <= 4; w++)
	    for (t = 1; t <= 4; t++) {
	      padrao[0] = i2c(i);
	      padrao[1] = i2c(j);
	      padrao[2] = i2c(k);
	      padrao[3] = i2c(w);
	      padrao[4] = i2c(t);
	      padrao[5] = '\0';
	      if (opcao == 1)
		sprintf(nome_arq2, "_%s.tmp", padrao);
	      else
		sprintf(nome_arq2, "~%s.tmp", padrao);
	      arq2 = fopen(nome_arq2, "rb");
	      if (arq2 != NULL) {
		if (ProcuraLinhaDist(nome_arq2, -1)) {
		  quant++;
		  fprintf(arq, DIV_linha_dist_min);
		}
		fclose(arq2);
	      }
	      else if (ProcuraLinhaPadrao(DIV_NOME_ARQ, padrao, opcao)) {
		quant++;
		fprintf(arq, DIV_linha_dist_min);
	      }
	    }
  }
  fclose(arq);
  return quant;
}

/* ---------------------------------------------------------------------- */

void ApagaArqTemp(int nivel, int opcao) {

/* Apaga arquivos temporarios. Se opcao == 1, e' usado o dominio de F
   como referencia; se opcao == 2, e' usado o dominio de G. */

  int i, j, k, w, t;
  char padrao[100], nome_arq[100];

  if (nivel < 1) nivel = 1; else if (nivel > 5) nivel = 5;

  if (nivel == 1) {
    for (i = 1; i <= 4; i++) {
      padrao[0] = i2c(i);
      padrao[1] = '\0';
      if (opcao == 1)
	sprintf(nome_arq, "_%s.tmp", padrao);
      else
	sprintf(nome_arq, "~%s.tmp", padrao);
      remove(nome_arq);
    }

  } else if (nivel == 2) {
    for (i = 1; i <= 4; i++)
      for (j = 1; j <= 4; j++) {
	padrao[0] = i2c(i);
	padrao[1] = i2c(j);
	padrao[2] = '\0';
	if (opcao == 1)
	  sprintf(nome_arq, "_%s.tmp", padrao);
	else
	  sprintf(nome_arq, "~%s.tmp", padrao);
	remove(nome_arq);
      }

  } else if (nivel == 3) {
    for (i = 1; i <= 4; i++)
      for (j = 1; j <= 4; j++)
	for (k = 1; k <= 4; k++) {
	  padrao[0] = i2c(i);
	  padrao[1] = i2c(j);
	  padrao[2] = i2c(k);
	  padrao[3] = '\0';
	  if (opcao == 1)
	    sprintf(nome_arq, "_%s.tmp", padrao);
	  else
	    sprintf(nome_arq, "~%s.tmp", padrao);
	  remove(nome_arq);
	}

  } else if (nivel == 4) {
    for (i = 1; i <= 4; i++)
      for (j = 1; j <= 4; j++)
	for (k = 1; k <= 4; k++)
	  for (w = 1; w <= 4; w++) {
	    padrao[0] = i2c(i);
	    padrao[1] = i2c(j);
	    padrao[2] = i2c(k);
	    padrao[3] = i2c(w);
	    padrao[4] = '\0';
	    if (opcao == 1)
	      sprintf(nome_arq, "_%s.tmp", padrao);
	    else
	      sprintf(nome_arq, "~%s.tmp", padrao);
	    remove(nome_arq);
	}

  } else if (nivel == 5) {
    for (i = 1; i <= 4; i++)
      for (j = 1; j <= 4; j++)
	for (k = 1; k <= 4; k++)
	  for (w = 1; w <= 4; w++)
	    for (t = 1; t <= 4; t++) {
	      padrao[0] = i2c(i);
	      padrao[1] = i2c(j);
	      padrao[2] = i2c(k);
	      padrao[3] = i2c(w);
	      padrao[4] = i2c(t);
	      padrao[5] = '\0';
	      if (opcao == 1)
		sprintf(nome_arq, "_%s.tmp", padrao);
	      else
		sprintf(nome_arq, "~%s.tmp", padrao);
	      remove(nome_arq);
	    }
  }
}

/* ---------------------------------------------------------------------- */

long SubdivisaoRecursiva(char *linha1, char *linha2, int nivel,
	       int nivel_maximo, retangulo A, retangulo B,
     	       double (*f1)(double, double), double (*f2)(double, double),
	       double (*f3)(double, double), double (*g1)(double, double),
	       double (*g2)(double, double), double (*g3)(double, double)) {

  int i, r[17], s[17], n_subdiv;
  char linha_atual1[100], linha_atual2[100];
  retangulo A2, B2;
  double x1, y1, x2, y2;
  static long quant = 0;

  if (nivel == 1) n_subdiv = N_SUBDIV1;
  else if (nivel == 2) n_subdiv = N_SUBDIV2;
  else if (nivel == 3) n_subdiv = N_SUBDIV3;
  else if (nivel == 4) n_subdiv = N_SUBDIV4;
  else n_subdiv = N_SUBDIV5;

  InterceptaCaixas(r, s, A, B, n_subdiv, f1, f2, f3, g1, g2, g3);

  i = 0;
  while (r[i]) {
    A2 = SelecionaRetangulo(r[i], A);
    B2 = SelecionaRetangulo(s[i], B);
    sprintf(linha_atual1, "%s%d", linha1, r[i]);
    sprintf(linha_atual2, "%s%d", linha2, s[i]);
    if (nivel < nivel_maximo) {
      SubdivisaoRecursiva(linha_atual1, linha_atual2, nivel + 1,
			     nivel_maximo, A2, B2, f1, f2, f3, g1, g2, g3);
    } else {
      quant++;
      if (DIV_RASTREAR)
	printf("\n%s %s", linha_atual1, linha_atual2);
      if (DIV_ARQ_DOMINIO)
	if (quant > 1) fprintf(DIV_arq_dados, "\n");
      PontoEmRetangulo(linha_atual1, DIV_Ret1_copia, &x1, &y1);
      PontoEmRetangulo(linha_atual2, DIV_Ret2_copia, &x2, &y2);
      fprintf(DIV_arq_dados, "\n%s %s [%6.6f]", linha_atual1, linha_atual2,
      DISTANCIA(f1(x1,y1),f2(x1,y1),f3(x1,y1),g1(x2,y2),g2(x2,y2),g3(x2,y2)));
      #ifdef COORD_CARTESIANAS
      fprintf(DIV_arq_dados," {(%6.6f, %6.6f), (%6.6f, %6.6f)}", x1,y1,x2,y2);
      #endif

      if (DIV_ARQ_DOMINIO)
	fprintf(DIV_arq_dados, "\n(%6.6f, %6.6f)", x1, y1);
      if (DIV_ARQ_VALORES)
	fprintf(DIV_arq_dados, " (%6.6f, %6.6f, %6.6f)", f1(x1, y1),
						    f2(x1, y1), f3(x1, y1));
      if (DIV_ARQ_DOMINIO)
	fprintf(DIV_arq_dados, "\n(%6.6f, %6.6f)", x2, y2);
      if (DIV_ARQ_VALORES)
      fprintf(DIV_arq_dados, " (%6.6f, %6.6f, %6.6f)", g1(x2, y2),
					    g2(x2, y2), g3(x2, y2));
    }
    i++;
  }

  return quant;
}

/* ---------------------------------------------------------------------- */

long Subdivisao(int nivel_max, retangulo Ret1, retangulo Ret2,
	       double (*f1)(double, double), double (*f2)(double, double),
	       double (*f3)(double, double), double (*g1)(double, double),
	       double (*g2)(double, double), double (*g3)(double, double)) {

/* Faz a subdivisao dos dominios A e B das superficies F e G ate' o nivel
   maximo nivel_max. Todo o resultado e' salvo no arquivo DIV_NOME_ARQ. E'
   retornado a quantidade de pontos encontrados. */

  long total;
  char linha1[100] = "", linha2[100] = "";

  DIV_Ret1_copia = Ret1, DIV_Ret2_copia = Ret2;
  if ((DIV_arq_dados = fopen(DIV_NOME_ARQ, "wt")) == NULL) {
    sprintf(mens_erro, "\n\nArquivo %s nao pode ser aberto", DIV_NOME_ARQ);
    Erro(mens_erro);
    Erro("\nERRO: no calculo da funcao Subdivisao()");
    return 0;
  }

  total = SubdivisaoRecursiva(linha1, linha2, 1, nivel_max,
				 Ret1, Ret2, f1, f2, f3, g1, g2, g3);
  fclose(DIV_arq_dados);
  return total;
}

/* ---------------------------------------------------------------------- */

/*** FIM DE SUBDIV.C ***/
